#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>

#define ID "Pktdate Rev. 1.4"

/* ============================== Prototypes ============================ */

static int init(int argc, char **argv, int require_arguments);
void log(int loglevel, const char *fmt, ...);
static int process_pkt(const char *filename);

#define LOG_ERROR   1
#define LOG_WARNING 2
#define LOG_NORMAL  3
#define LOG_DEBUG   4

const char *log_chars = " ?! d";

typedef unsigned long flag_t;  /* for at least 32 bit flags */

                                /* constants that describe errors    */
#define FTSC_FLAWY  1           /* FTSC field has correctable errors */
#define FTSC_BROKEN 2           /* FTSC field can't even be parsed   */
#define FTSC_SPILLED_ZERO 4     /* FTSC field too long -> msg looks  */
                                /* like it has a zero length subject */
#define UNPLAUSIBLE_DATE 8      /* date not within defined tolerance */
                                /* boundaries                        */
#define FTSC_SEADOG 16          /* Seadog style string in the FTSC   */
                                /* date field                        */
#define UD_SQUISH_BUG 32        /* Unplausible date: 7 bit overflow  */
#define UD_YEAR_DISP  64        /* Only year is wrong                */
#define FTSC_TS_BROKEN 128      /* Only timestamp broken, date is OK */
#define UD_FIELD_UNINITIALISED 256 /* date field uninitialised       */
#define FATAL_ERROR ((unsigned long)-1L)

/* =========================== Global variables ========================= */

struct _config
{
    int fix : 1;               /* -c : try to correct errors       */
    int fixseadog : 1;         /* -S : fix Seadog style date field */
    int qqq : 1;               /* -q : accept Fastecho QQQ files   */
    int keep: 1;               /* -k : keep copies of broken files */
    const char *logfilename;   /* -L : log file name               */
    int   maxlog;              /* -l : maximum log level           */
    unsigned long futureTol;   /* -f : Tolerance for date in days  */
                               /*      in future direction         */
    unsigned long pastTol;     /* -p : Tolerance into the past     */
    unsigned long sfutureTol;  /* -F : for speculative error corr. */
    unsigned long spastTol;    /* -P : for speculative error corr. */

    FILE *logfile;
} config;


/* ============================= Main Program =========================== */

int main(int argc, char **argv)
{
    int i;
    int rv=0, rvt;

    for (i = init (argc, argv, 1); i < argc; i++)
    {
        rvt = process_pkt(argv[i]);
        if (rvt > rv)
            rv = rvt;
    }

    return rv;
}


/* ======================== Process a PKT file ========================== */

struct msginpkt
{
    long absoffset;      /* ftell() result at message start in pkt file */
    long reloffsets[4];  /* From, To, Subject, Body */
    long lengths[4];     /* byte lengths including terminating zeros */

    char date_string[22];
    char new_date_string[22];
    flag_t needs_fix;
};

struct pktfile
{
    FILE *f; const char *filename;
    int day, month, year;
    unsigned int capability_word;
    size_t nmsgs, nmaxmsgs;
    struct msginpkt *msg;
    int is_qqq; /* if 1: do not write trailing 00 00 when fixing up! */
    int keep_me; /* if 1: keep a copy of the original corrupted file */

    flag_t needs_fix;
};

static int  read_pkt  (struct pktfile *ppkt, FILE *f, const char *filename);
static void fix_pkt   (struct pktfile *ppkt);
static void show_pkt  (struct pktfile *ppkt);
static int  write_pkt (struct pktfile *ppkt);
int movefile (const char *old_name, const char *new_name);
int fexist (const char *filename);
long daydiff(struct tm*, struct tm*);

static int process_pkt (const char *filename)
{
    FILE *f;
    struct pktfile pkt;

    if ((f = fopen(filename, "rb")) == NULL)
    {
        log (LOG_ERROR, "Cannot open file %s. Skipping.\n", filename);
        return FATAL_ERROR;
    }

    if (read_pkt(&pkt, f, filename) == -1)
    {
        if (pkt.f   != NULL) fclose(f);
        if (pkt.msg != NULL) free(pkt.msg);
        return FATAL_ERROR;
    }

    fix_pkt(&pkt);
    show_pkt(&pkt);

    if (pkt.needs_fix && config.fix)
    {
        if (config.keep)
            pkt.keep_me = 1;
        if (!write_pkt(&pkt))
        {
            log (LOG_NORMAL, "Wrote changes to %s.", pkt.filename);
        }
    }


    if (pkt.f != NULL) fclose(f);
    free(pkt.msg);
    return (int)pkt.needs_fix;
}

static int read_pkt(struct pktfile *ppkt, FILE *f, const char *filename)
{
    unsigned char buffer[0x3A];
    long ofs, safe_ofs;
    int i, spz;

    ppkt->f = f;
    ppkt->filename = filename;
    ppkt->nmaxmsgs = 0; ppkt->nmsgs = 0; ppkt->msg = NULL;
    ppkt->needs_fix = 0;
    ppkt->is_qqq = 0;
    ppkt->keep_me = 0;

    i = fread(buffer, 1, ofs = 0x3A, f);

    if (i != 0x3A)
    {
        if (i == 0 || feof(f))
        {
           log (LOG_WARNING, "Zero length file %s.\n", ppkt->filename);
           return -1;
        }

        log (LOG_ERROR, "Read error on file %s.\n", ppkt->filename);
        return -1;
    }

    ppkt->day   = (buffer[9]<<8) + buffer[8];
    ppkt->month = (buffer[7]<<8) + buffer[6] + 1;
    ppkt->year  = (buffer[5]<<8) + buffer[4];

    ppkt->capability_word = buffer[0x2C] + (buffer[0x2D] << 8);
    if (ppkt->capability_word != (buffer[0x28] << 8) + buffer[0x29])
    {
        ppkt->capability_word = 0;  /* stone age! */
    }

    if (ppkt->capability_word != 0 && ppkt->capability_word != 1)
    {
        log (LOG_ERROR,  "Skipped %s (unknown packet type).", ppkt->filename);
        return -1;
    }

    while (1)
    {
        if (fread(buffer,1,2,f) != 2)
        {
            if(config.qqq)
            {
                ppkt->is_qqq = 1;
                break;  /* end of FE QQQ file */
            }
            else
            {
                log (LOG_ERROR, "%s: Unexpected end of file. Skipping.\n",
                     ppkt->filename);
                return -1;
            }
        }
        ofs += 2;

        if ((buffer[0] == buffer[1]) && (buffer[1] == 0))  /* end of pkt */
        {
            break;
        }
        if ((buffer[0] != 2) || (buffer[1] != 0))          /* bad pkt */
        {
            safe_ofs = ofs - 2;

            if (ppkt->nmsgs > 0 && ppkt->msg[ppkt->nmsgs-1].lengths[0] == 1)
            {
                /* Speculative read ahead; maybe an overflowing date field
                   has misled us to assume an empty "to" field, so we seek
                   to the next null and see if another message starts after
                   that one. This is the most complicated code of the whole
                   program ...*/

                while (buffer[0] != 0 && (!feof(f)))
                {
                    buffer[0] = buffer[1];
                    buffer[1] = fgetc(f);
                    ofs++;
                }

                spz = -1;  /* no sucess by default */

                if (feof(f) && config.qqq && buffer[0] == 0)
                {
                    spz = 1;  /* reached end of a QQQ file */
                    ppkt->is_qqq = 1;
                }
                else if (buffer[0] == 0)
                {             /* reached end of this message */
                    buffer[0] = buffer[1]; buffer[1] = fgetc(f); ofs++;
                    if ((buffer[0] == 2 || buffer[0] == 0) && buffer[1] == 0)
                    {
                        spz = 2;
                    }
                }

                if (spz != -1)
                {
                    /* bingo! we found the end of the last message */

                    ppkt->msg[ppkt->nmsgs - 1].needs_fix |=
                        FTSC_SPILLED_ZERO;
                    ppkt->needs_fix |= FTSC_SPILLED_ZERO;

                    for (i = 0; i < 3; i++)
                    {
                        ppkt->msg[ppkt->nmsgs - 1].reloffsets[i] =
                            ppkt->msg[ppkt->nmsgs - 1].reloffsets[i + 1];
                        ppkt->msg[ppkt->nmsgs - 1].lengths[i] =
                            ppkt->msg[ppkt->nmsgs - 1].lengths[i + 1];
                    }

                    ppkt->msg[ppkt->nmsgs - 1].reloffsets[3] =
                        safe_ofs - ppkt->msg[ppkt->nmsgs - 1].absoffset;
                    ppkt->msg[ppkt->nmsgs - 1].lengths[3] =
                        ofs - spz - safe_ofs;

                    if (buffer[0] == 0 || (feof(f) && config.qqq))
                    {
                        break;
                    }
                }
            }
            if ( ppkt->nmsgs < 1 ||
                 (!(ppkt->msg[ppkt->nmsgs - 1].needs_fix & FTSC_SPILLED_ZERO))
               )
            {
                log (LOG_ERROR, "%s: Packed message is corrupt.\n",
                     ppkt->filename);
                return -1;
            }
        }

        if (fread(buffer+2,1,0x20,f) != 0x20)
        {
            log (LOG_ERROR, "%s: Unexpected end of file. Skipping.\n",
                 ppkt->filename);
            return -1;
        }
        ofs += 0x20;

        if (ppkt->nmaxmsgs <= ppkt->nmsgs)
        {
            if (ppkt->msg == NULL)
            {
                if ((ppkt->msg = malloc(sizeof(struct msginpkt) *
                                        (ppkt->nmaxmsgs = 10))) == NULL)
                {
                    log (LOG_ERROR, "Out of memory!\n");
                    return -1;
                }
            }
            else
            {
                if ((ppkt->msg = realloc(ppkt->msg,
                                         sizeof(struct msginpkt) *
                                         (ppkt->nmaxmsgs *= 2))) == NULL)
                {
                    log (LOG_ERROR, "Out of memory!\n");
                    return -1;
                }
            }
        }

        ppkt->msg[ppkt->nmsgs].absoffset = ofs - 0x22;
        ppkt->msg[ppkt->nmsgs].needs_fix = 0;

        memcpy(ppkt->msg[ppkt->nmsgs].date_string, buffer+0xE, 20);
        ppkt->msg[ppkt->nmsgs].date_string[20] = '\0';

        for (i=0;i<4;i++)
        {
            int c;
            ppkt->msg[ppkt->nmsgs].reloffsets[i] = ofs -
                ppkt->msg[ppkt->nmsgs].absoffset;

            while ((c=fgetc(f)) != 0)
            {
                ofs++;
                if (c==EOF)
                {
                    log (LOG_ERROR, "%s: Unexpected end of file. Skipping.\n",
                         ppkt->filename);
                    return -1;
                }
            }
            ofs++;

            ppkt->msg[ppkt->nmsgs].lengths[i] =
                ofs - (ppkt->msg[ppkt->nmsgs].reloffsets[i] +
                       ppkt->msg[ppkt->nmsgs].absoffset);
        }
        ppkt->nmsgs ++;

    }

    return 0;
}

#define BUFFER_SIZE 32768       /* Any system should be able to malloc 32K, and
                                   it gives tolerable good performance */

static int  write_pkt (struct pktfile *ppkt)
{
    char *tmp_file_name;
    char *y2k_file_name;
    FILE *tmp_file;
    unsigned char *buffer;
    size_t i,k;
    struct msginpkt *pmsg;
    int j;
    long n;
    size_t toread;

    if ((buffer = malloc(BUFFER_SIZE)) == NULL)
    {
        log (LOG_ERROR, "Out of memory.");
        return -1;
    }

    if (((tmp_file_name = tmpnam(NULL)) == NULL) ||
        ((tmp_file = fopen(tmp_file_name, "wb")) == NULL))
    {
        log (LOG_ERROR, "Could not create temporary file - fix aborted.");
        free(buffer);
        return -1;
    }

    /* transfer the pkt header */
    if (fseek(ppkt->f, 0, SEEK_SET) != 0) goto read_error;
    if (fread(buffer, 0x3A, 1, ppkt->f) != 1) goto read_error;
    if (fwrite(buffer, 0x3A, 1, tmp_file) != 1) goto write_error;

    for (i=0; i < ppkt->nmsgs; i++)
    {
        pmsg = ppkt->msg + i;

        /* read msg header, fix if necessary, and save it again */
        if (fseek(ppkt->f, pmsg->absoffset, SEEK_SET) != 0) goto read_error;
        if (fread(buffer, 0x22, 1, ppkt->f) != 1) goto read_error;
        if (buffer[0] != 2 || buffer[1] != 0) goto integrity_error;
        if (pmsg->needs_fix)
        {
            memcpy(buffer + 0xe, pmsg->new_date_string, 20);
        }
        if (fwrite(buffer, 0x22, 1, tmp_file) != 1) goto write_error;

        for (j = 0; j < 4; j++)
        {
            if (fseek(ppkt->f, pmsg->absoffset + pmsg->reloffsets[j], SEEK_SET)
                != 0) goto read_error;

            for (n = 0; n < pmsg->lengths[j]; n += (long) toread)
            {
                toread = (n + BUFFER_SIZE < pmsg->lengths[j])
                    ? BUFFER_SIZE  : (size_t)(pmsg->lengths[j] - n);

                if (fread(buffer, toread, 1, ppkt->f) != 1) goto read_error;
                if (fwrite(buffer, toread, 1, tmp_file) != 1) goto write_error;
            }
        }
    }

    /* write 00 00, unless it is a Fastecho .QQQ file */
    if (!ppkt->is_qqq)
    {
        buffer[0] = 0; buffer[1] = 0;
        if (fwrite(buffer, 2, 1, tmp_file) != 1) goto write_error;
    }

    fclose(tmp_file);
    free(buffer);
    fclose(ppkt->f); ppkt->f = NULL;

    if (ppkt->keep_me)
    {
        /* we want to keep a copy of the file renamed to .y2k */

        i = strlen(ppkt->filename);
        if (!i || (y2k_file_name = malloc(i + 5)) == NULL)
        {
            log (LOG_ERROR, "Out of memory - cannot keep original file.");
            remove(ppkt->filename);
        }
        else
        {
            /* construct the new file name */

            memmove(y2k_file_name, ppkt->filename, i + 1);
            for (k = i - 1; k > 0 && y2k_file_name[k] != '.'
                 && y2k_file_name[k] != '/' && y2k_file_name[k] != '\\'
                 && y2k_file_name[k] != ':'; k--);
            if (y2k_file_name[k] != '.')
            {
                k = i;
            }
            /* now k is the position where we can put the new extension */

            /* try the default extension first */
            strcpy(y2k_file_name + k,  ".y2k");

            /* but if it exists, try, .Y## extensions */
            for (i = 1; fexist(y2k_file_name) && i < 100; i++)
            {
                sprintf (y2k_file_name + k, ".y%02d", i);
            }

            if (i == 100) /* we can't fine any free name - TROUBLE! */
            {
                log (LOG_ERROR, "Cannot find a file name to keep a copy of the corrupted PKT file.");
                remove(ppkt->filename);
            }
            else
            {
                log (LOG_WARNING, "Keeping a copy of the corrupted PKT as %s for manual inspection.", y2k_file_name);
                if (movefile(ppkt->filename, y2k_file_name) == -1)
                {
                    log(LOG_ERROR, "Cannot move %s to %s.",
                        ppkt->filename, y2k_file_name);
                    remove(ppkt->filename);
                }
            }
        }
    }
    else
    {
        remove(ppkt->filename);
    }

    /* move the new file in place */;
    movefile(tmp_file_name, ppkt->filename);
    return 0;

 write_error:
    log (LOG_ERROR, "Write error on temporary file %s.\n", tmpnam);
    fclose(tmp_file);
    remove(tmp_file_name);
    free(buffer);
    return -1;

 integrity_error:
    log (LOG_ERROR, "Program / data integrity error.", tmpnam);
    fclose(tmp_file);
    remove(tmp_file_name);
    free(buffer);
    return -1;

 read_error:
    log (LOG_ERROR, "Read error on file %s.\n", ppkt->filename);
    fclose(tmp_file);
    remove(tmp_file_name);
    free(buffer);
    return -1;
}

static flag_t parse_ftsc_date(struct tm *, struct tm *, char *);
static void   make_ftsc_date (char *, const struct tm *);

static void fix_pkt  (struct pktfile *ppkt)
{
    struct tm tm, tms;
    struct tm *pnow;
    time_t t_now;
    struct msginpkt *pmsg;
    size_t i;
    int s;
    int currentreason;
    int parseresult;

    time(&t_now);
    pnow = localtime(&t_now);   /* get the current time */

    for (i = 1; i <= ppkt->nmsgs; i++)
    {
        pmsg = ppkt->msg + i - 1;

                                /* parse the FTSC date field */
        memcpy(pmsg->new_date_string, pmsg->date_string,
               sizeof(pmsg->new_date_string));


        parseresult = parse_ftsc_date(&tm, pnow, pmsg->new_date_string);

        pmsg->needs_fix |= parseresult;

        if (parseresult & FTSC_BROKEN) /* parsing failed fatally */
        {
            memcpy(&tm, pnow, sizeof(struct tm));
        } else
        {
            /* date plausibility checks  */
            if (daydiff(&tm, pnow) > (long)config.futureTol ||
                daydiff(pnow, &tm) > (long)config.pastTol)
            {
                pmsg->needs_fix |= UNPLAUSIBLE_DATE;

                /* Speculative error correction */

                for (s = 1; s <= 6; s++)
                {
                    memcpy(&tms, &tm, sizeof(struct tm));

                    switch (s)
                    {
                    case 1:
                        if (tm.tm_year == 70 && tm.tm_mon == 0 &&
                            tm.tm_mday == 1 && tm.tm_hour == 0 &&
                            tm.tm_min == 0 && tm.tm_sec == 0)
                        {
                            /* uninitialised date field */
                            memcpy(&tms, pnow, sizeof(struct tm));
                            currentreason = UD_FIELD_UNINITIALISED;
                        }
                        break;

                    case 2:

                        /* #1: Try to undo the 7 bit overflow bug (happens
                           when 00 is interpreted as "1900" and then a
                           timestamp structure with 7 bit year offset
                           (rel. 1980) is filled in. Observed in Squish. */

                        tms.tm_year = tm.tm_year - 28;
                        currentreason = UD_SQUISH_BUG;
                        break;

                    case 3:

                        /* #2..#4: Assume the sysop has just put his clock back
                           by some or several years */

                        tms.tm_year = pnow->tm_year;
                        currentreason = UD_YEAR_DISP;
                        break;

                    case 4:
                        tms.tm_year = pnow->tm_year - 1;
                        currentreason = UD_YEAR_DISP;
                        break;

                    case 5:
                        tms.tm_year = pnow->tm_year + 1;
                        currentreason = UD_YEAR_DISP;
                        break;

                    case 6:
                        /* no plan - set it to current time */
                        memcpy(&tms, pnow, sizeof(struct tm));
                        currentreason = 0;
                        break;
                    }

                    if (daydiff(&tms, pnow) < (long)config.sfutureTol &&
                        daydiff(pnow, &tms) < (long)config.spastTol)
                    {
                        memcpy(&tm, &tms, sizeof(struct tm));
                        pmsg->needs_fix |= currentreason;
                        break; /* leave the loop */
                    }
                }
            }
        }

        if (pmsg->needs_fix)
        {
            make_ftsc_date(pmsg->new_date_string, &tm);
            ppkt->needs_fix |= pmsg->needs_fix;
        }
    }
}

static void show_pkt(struct pktfile *ppkt)
{
    int i;
    struct msginpkt *pmsg;

    log ((ppkt->needs_fix) ? LOG_WARNING : LOG_NORMAL,
         "%s processing %s (type: %s)",
         ID,
         ppkt->filename,
         ppkt->capability_word == 1 ? "2+" : "stone-aged");

    log ((ppkt->needs_fix) ? LOG_WARNING : LOG_NORMAL,
         "Packet header date (D/M/Y): %d.%d.%d",
         ppkt->day, ppkt->month, ppkt->year);

    for (i = 1; i <= ppkt->nmsgs; i++)
    {
        pmsg = ppkt->msg + i - 1;

        if (pmsg->needs_fix && config.fix)
        {
            log (LOG_WARNING,
                 "Message #%03d FTSC date: \"%s\" fixed (reason %lu) to \"%s\"",
                 i, pmsg->date_string,
                 (unsigned long) pmsg->needs_fix, pmsg->new_date_string);
        }
        else if (pmsg->needs_fix)
        {
            log (LOG_WARNING,
                 "Message #%03d FTSC date: \"%s\" needs fix (reason %lu).",
                 i, pmsg->date_string, (unsigned long) pmsg->needs_fix);
        }
        else
        {
            log (LOG_NORMAL, "Message #%03d FTSC date: \"%s\" is OK.\n", i,
                 pmsg->date_string);
        }
    }
}


/* ======================== general utility routines ==================== */


/* Some toupper routines crash when they get invalid input. As this program
   is intended to be portable and deal with any sort of malformed input,
   we have to provide our own toupper routine. */
char safe_toupper(char c)
{
    const char *from_table = "abcdefghijklmnopqrstuvwxyz";
    const char *to_table   = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    const char *p;

    if ((p = strchr(from_table, c)) != NULL)
    {
        return (to_table[p - from_table]);
    }

    return c;
}


int fexist(const char *filename)
{
    FILE *f = fopen(filename, "rb");

    if (f != NULL)
    {
        fclose(f);
        return 1;
    }
    else
    {
        /* at least we don't have write access to the file - this is good */
        /* enough for our purposes: we cannot destroy anything at least.  */
        return 0;
    }
}

#define MOVE_FILE_BUFFER_SIZE 16384

int movefile(const char *from, const char *to)
{
    int rc;
    char *buffer;
    size_t read;
    FILE *fin, *fout;


    rc = rename(from, to);
    if (!rc) return 0;        /* rename succeeded. fine! */

    /* Rename did not succeed, probably because the move is accross
       file system boundaries. We have to copy the file. */

    buffer = malloc(MOVE_FILE_BUFFER_SIZE);
    if (buffer == NULL) return -1;

    fin = fopen(from, "rb");
    if (fin == NULL) { free(buffer); return -1; }

    fout = fopen(to, "wb");
    if (fout == NULL) { free(buffer); return -1; }

    while ((read = fread(buffer, 1, MOVE_FILE_BUFFER_SIZE, fin)) > 0)
    {
    if (fwrite(buffer, 1, read, fout) != read)
    {
        fclose(fout); fclose(fin); remove(to); free(buffer);
        return -1;
    }
    }

    if (ferror(fout) || ferror(fin))
    {
    fclose(fout);
    fclose(fin);
    free(buffer);
    remove(to);
    return -1;
    }

    fclose(fout);
    fclose(fin);
    free(buffer);
    remove(from);
    return 0;
}


/* ======================== date utility routines ======================= */


/* The Julian Date is a unique day number. It can be used to calculate the */
/* difference in days between dates, or to calculate the day of week.      */
/* Works from May 1600 onwards. */

long get_julian_date(int day, int month, int year)
{
    if (month > 2)
    {
        month -= 3;
    }
    else
    {
        month += 9;
        year--;
    }
    year -= 1600;
    return (((((long)year / 100)  *146097L) / 4) +
            ((((long)year % 100)  *1461) / 4) +
            (((153 * (long)month) + 2) / 5) + day + 59);
}

long daydiff(struct tm* first, struct tm* second)
{
    long diff =
           - get_julian_date(second->tm_mday,
                             second->tm_mon + 1,
                             second->tm_year + 1900)
           + get_julian_date(first->tm_mday,
                             first->tm_mon + 1,
                             first->tm_year + 1900);
    return diff;
}



static const char *months[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun",
                               "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

int get_month(const char *pmon, flag_t *flag)
{
    int i;

    if (strlen(pmon) != 3 && flag != NULL)
    {
        (*flag) |= FTSC_FLAWY;
    }

    for (i=0; i < 12; i++)
    {
        if (pmon[0] == months[i][0] &&
            pmon[1] == months[i][1] &&
            pmon[2] == months[i][2])
        {
            return i;
        }
    }

    for (i=0; i < 12; i++)
    {
        if (safe_toupper(pmon[0]) == safe_toupper(months[i][0]) &&
            safe_toupper(pmon[1]) == safe_toupper(months[i][1]) &&
            safe_toupper(pmon[2]) == safe_toupper(months[i][2]))
        {
            (*flag) |= FTSC_FLAWY;
            return i;
        }
    }

    (*flag) |= FTSC_BROKEN;
    return 0;
}


static flag_t parse_ftsc_date(struct tm * ptm, struct tm * pnow,
                              char *pdatestr)
{
    const char *pday, *pmon, *pyear, *phour, *pminute, *psecond;
    flag_t rval;
    char buf[22];

    pday = pmon = pyear = phour = pminute = psecond = NULL;

    rval = FTSC_BROKEN;

    memcpy(buf, pdatestr, 21); buf[21] = 0;

    if ((pday = strtok(pdatestr, " ")) != NULL)
        if ((pmon = strtok(NULL, " ")) != NULL)
            if ((pyear = strtok(NULL, " ")) != NULL)
                if ((phour = strtok(NULL, ":")) != NULL)
                    if ((pminute = strtok(NULL, ":")) != NULL)
                        if ((psecond = strtok(NULL, " ")) != NULL)
                            rval = 0;

    if (rval == FTSC_BROKEN)
    {
                 /* let's try and see if it might be the old SeaDog format */

        rval = FTSC_BROKEN;

        if ((strtok(buf, " ")) != NULL)
            if ((pday = strtok(NULL, " ")) != NULL)
                if ((pmon = strtok(NULL, " ")) != NULL)
                    if ((pyear = strtok(NULL, " ")) != NULL)
                        if ((phour = strtok(NULL, ": ")) != NULL)
                            if ((pminute = strtok(NULL, ": ")) != NULL)
                            {
                                psecond = NULL;
                                if (config.fixseadog)
                                    rval = FTSC_SEADOG;
                                else
                                    rval = 0;
                            }
    }


    ptm->tm_sec = ptm->tm_min = ptm->tm_hour = ptm->tm_mday = ptm->tm_mon =
        ptm->tm_year = 0;

    while (rval != FTSC_BROKEN)    /* at least we could tokenize it! */
    {
        if (psecond != NULL)
        {
            ptm->tm_sec = atoi(psecond);   /* Is the number of seconds valid? */
            if (ptm->tm_sec < 0 || ptm->tm_sec > 59)
            {
                rval |= FTSC_TS_BROKEN; ptm->tm_sec = 0;
            }
            if (strlen(psecond) == 1)
            {
                rval |= FTSC_FLAWY; ptm->tm_sec *= 10;
            }
        }
        else
        {
            ptm->tm_sec = 0;
        }

        ptm->tm_min = atoi(pminute);   /* Is the number of minutes valid? */
        if (ptm->tm_min < 0 || ptm->tm_min > 59)
        {
            rval |= FTSC_TS_BROKEN; ptm->tm_min = 0;
        }

        ptm->tm_hour = atoi(phour);    /* Is the number of hours valid? */
        if (ptm->tm_hour < 0 || ptm->tm_hour>23)
        {
            rval |= FTSC_TS_BROKEN; ptm->tm_hour = 0;
        }

        ptm->tm_mday = atoi(pday);     /* Is the day in the month valid? */
        if (ptm->tm_mday < 1 || ptm->tm_mday>31) { rval |= FTSC_BROKEN; break; }

        ptm->tm_mon = get_month(pmon, &rval); /* Is the month valid? */

        if (strlen(pyear) != 2)        /* year field format check */
        {
            rval |= FTSC_FLAWY;
        }

        if (*pyear)
        {
            ptm->tm_year = (*pyear - '0'); /* allows for the ":0" bug */
            for (pyear++; isdigit(*pyear); pyear++)
            {
                ptm->tm_year *= 10;
                ptm->tm_year += (*pyear - '0');
            }
            if (*pyear)
            {
                rval |= FTSC_BROKEN;
                break;
            }
        }
        else
        {
            rval |= FTSC_BROKEN;
            break;
        }

        if (ptm->tm_year < 100)  /* correct date field! */
        {
            while (pnow->tm_year - ptm->tm_year > 50)
            {                         /* sliding window adaption */
                ptm->tm_year += 100;
            }
        }
        else if (ptm->tm_year < 1900) /* probably the field directly */
                                      /* contains tm_year, like produced */
                                      /* by the Timed/Netmgr bug and others */
        {
            rval |= FTSC_FLAWY;
        }
        else                          /* 4 digit year field, not correct! */
        {
            ptm->tm_year -= 1900;
            rval |= FTSC_FLAWY;
        }
        break;
    }


    return rval;

}

static void make_ftsc_date(char *pdate, const struct tm *ptm)
{
    sprintf(pdate, "%02d %-3.3s %02d  %02d:%02d:%02d",
            ptm->tm_mday % 100, months[ptm->tm_mon], ptm->tm_year % 100,
            ptm->tm_hour % 100, ptm->tm_min % 100, ptm->tm_sec % 100);
}


/* ========================== Routine for logging ======================= */
void log(int loglevel, const char *fmt, ...)
{
    va_list argptr;
    int neednl = fmt[strlen(fmt) - 1] != '\n';

    time_t t;
    struct tm *tm;
    char cpTime[22];
    static char *cpMonths[]={"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
                             "Aug", "Sep", "Oct", "Nov", "Dec", "???"};

    va_start(argptr, fmt);
    if (config.maxlog >= loglevel)
    {
        time(&t);
        tm=localtime(&t);
        sprintf(cpTime,"%c %02d %-3.3s %02d:%02d:%02d PKTD ",
                log_chars[loglevel],
                tm->tm_mday, cpMonths[tm->tm_mon], tm->tm_hour, tm->tm_min,
                tm->tm_sec);

        if (config.logfile != NULL)
        {
            fprintf(config.logfile, "%s", cpTime);
            vfprintf(config.logfile, fmt, argptr);
            if (neednl) fprintf(config.logfile, "\n");
        }
        vfprintf(stdout, fmt, argptr);
        if (neednl) fprintf(stdout, "\n");
    }
}



/* ===================== Command line analysator ======================== */
/* Parses a command line according to the POSIX standards.                */
/* Option clustering (-abc instead of -a -b -c) is not supported.         */

static const char *valid_options="?qcSkl:L:f:p:F:P:";
static const char *usage="\
Usage:\tpktdate [<option> ...] <filename> [<filename> ...]\n\n\
Where <option> can be:\n\
  -c            Try to correct broken date fields in the PKT file. (The \n\
                default is only to show those errors, but not to correct them).\n\
  -l<loglevel>  Set the maximum loglevel. <loglevel> can be:\n\
                1  only log fatal errors\n\
                2  log fatal errors and non-fatal errors (warnings)\n\
                3  log errors, warnings, and information messages (default).\n\
                4  also log debug information.\n\
  -L<filename>  Set the name and path of the log file. Default: no log file.\n\
  -f<tolerance> Set the future tolerance for date field checking. A message \n\
                date that is more than <tolerance> in the future will \n\
                be considered broken and, if -c is specified be replaced by \n\
                the current date. <tolerance> can be given in days, or \n\
                in months by using the sufffix \"m\" like in \"-f3m\", or in years\n\
                by using the sufffix \"y\". Default: 31 days.\n\
  -p<tolerance> Works like -f, but specifies the tolerance into the past.\n\
                Default: 364 days.\n\
  -P<tolerance> Sets the past tolerance time for speculative error correction.\n\
                Default: 1 month. See pktdate.txt for an explanation.\n\
  -k            When correcting errors, keep a copy of the original PKT.\n\
  -q            Accept Fastecho .QQQ style packet files.\n\
  -S            Change Seadog style date fields (like \"Thu  1 Apr 99 10:05\", a\n\
                format that is deprecated) to the new FTSC format.\
";

static void initoptions (void)
{
    config.fix = 0;
    config.logfilename = NULL;    /* "pktdate.log"; */
    config.maxlog = LOG_NORMAL;
    config.futureTol  = 31L;      /* One month into the future. */
    config.pastTol    = 364L;     /* Almost one year into the past. */
    config.sfutureTol = 1L;       /* Acceptance window for error correction */
    config.spastTol   = 31L;      /* Acceptance window for error correction */

}

static void finalizeconfig (void)
{
    if (config.logfilename != NULL)
    {
        config.logfile = fopen (config.logfilename, "a");
    }
    else
    {
        config.logfile = NULL;
    }
}

static void processoption (char option, char *argument)
{
    int flag;

    switch (option)
    {
    case 'c':
        config.fix = 1;
        break;

    case 'q':
        config.qqq = 1;
        break;

    case 'k':
        config.keep = 1;
        break;

    case 'S':
        config.fixseadog = 1;
        break;

    case 'L':
        config.logfilename = argument;
        break;

    case 'l':
        if (atoi(argument) != 0)
        {
            config.maxlog = atoi(argument);
        }
        break;

    case 'f':
    case 'p':
    case 'F':
    case 'P':
        if (*argument && safe_toupper(argument[strlen(argument) - 1]) == 'Y')
        {
            argument[strlen(argument) - 1] = '\0';
            flag = 365;
        }
        else
        if (*argument && safe_toupper(argument[strlen(argument) - 1]) == 'M')
        {
            flag = 31;
        }
        else
        {
            flag = 1;
        }
        if (atol(argument) != 0L)
        {
            if (option == 'f')
            {
                config.futureTol = atol(argument) * ((unsigned long)flag);
            }
            else if (option == 'p')
            {
                config.pastTol = atol(argument) * ((unsigned long) flag);
            }
            else if (option == 'F')
            {
                config.futureTol = atol(argument) * ((unsigned long) flag);
            }
            else if (option == 'P')
            {
                config.spastTol = atol(argument) * ((unsigned long) flag);
            }
        }
        break;

    default:
        if (option != '?')
        {
            fprintf (stderr, "Option \"-%c\" is not yet implemented.\n",
                     option);
        }
        else
        {
            printf ("%s, written by Tobias Ernst @ 2:2476/418\n\n", ID);
            if (argument != NULL)
            {
                fprintf (stderr, "Error: %s\n\n", argument);
            }
            printf ("%s\n", usage);
        }
        exit ((argument == NULL) ? 0 : 8);

    }
}

int init(int argc, char **argv, int require_arguments)
{
    int i, firstargument = 0;
    const char *cp;
    char errmsg[80];

    /* this sets all configuration parameters to their default values */
    initoptions();

    /* process all options. stop as soon as we find an argument.
       (options like -l1 MODIFY the behaviour of the program, arguments
       like filenames tell it what to do at all. Optoins come first, then come
       the filenames, optionally separated by a double dash */
    for (i = 1; i < argc && firstargument == 0; i++)
    {
        switch (argv[i][0])
        {
        case '-':
            switch (argv[i][1])
            {
            case '-':    /* -- indicates end of options -- */
                firstargument = i + 1;
                break;

            case '\0':   /* single dash is probably a file name argument */
                firstargument = i;
                break;

            default:                              /* some sort of option */
                cp = strchr(valid_options, argv[i][1]);
                if (cp == NULL)                   /* invalid option */
                {
                    sprintf (errmsg, "Unknown option \"-%c\".", argv[i][1]);
                    processoption ('?', errmsg); /* error */
                }
                else if (cp[1] == ':')            /* option with argument */
                {
                    if (argv[i][2] == '\0' && i + 1 < argc)
                    {
                        processoption (*cp, argv[i + 1]);
                        i++;
                    }
                    else if (argv[i][2] != '\0')
                    {
                        processoption (*cp, argv[i]+2);
                    }
                    else
                    {
                        sprintf (errmsg,
                                 "Option \"-%c\" requires an argument.",
                                 *cp);
                        processoption ('?', errmsg);
                    }
                }
                else                              /* option w/o argument */
                {
                    if (argv[i][2] != '\0')       /* error - unexp. opt arg */
                    {
                        sprintf (errmsg,
                                 "Parse error at command line argument #%d.",
                                 i);
                        processoption ('?', errmsg);
                    }
                    else
                    {
                        processoption (*cp, NULL);
                    }
                }
            }
            break;
        default:
            firstargument = i;
        }
    }

    if (firstargument == 0 && require_arguments)
    {
        sprintf (errmsg, "No arguments given.");
        processoption ('?', NULL);
    }

    finalizeconfig();

    return firstargument;
}
